// SPDX-License-Identifier: MIT
/*
 * Copyright (c) 2023 Atmark Techno,Inc.
 */

#include "device-info.h"
#include "security_imx8mp.h"

char *boot_cfg_lock_settings_imx8mp(uint32_t boot_cfg_lock_mem)
{
	if ((boot_cfg_lock_mem
	    & IMX8MP_BOOT_CFG_LOCK_MASK)
	    == IMX8MP_BOOT_CFG_LOCK_OPWP) {
		return "write protected";
	}

	return "open";
}

char *cst_srk_lock_settings_imx8mp(uint32_t status_mem)
{
	if ((status_mem
	    & IMX8MP_CST_SRK_LOCK_MASK)
	    == IMX8MP_CST_SRK_LOCK_ENABLE) {
		return "write protected";
	}

	return "open";
}

int print_fuse_status_lock_imx8mp(unsigned char *ocotp_data)
{
	uint32_t status_mem;
	char *status_str;

	memcpy(&status_mem,
		ocotp_data + IMX8MP_BOOT_CFG_LOCK_OFFSET, 4);

	status_str = boot_cfg_lock_settings_imx8mp(status_mem);
	PRINTF_CHECK("boot_cfg_lock %s\n", status_str);

	status_str = cst_srk_lock_settings_imx8mp(status_mem);
	PRINTF_CHECK("cst_srk_lock %s\n", status_str);

	return 0;
}

int is_secure_fuse_boot_cfg_lock_imx8mp(unsigned char *ocotp_data)
{
	if (secure_status_boot_cfg_lock_imx8mp(ocotp_data) == SECURE)
		return EXIT_SUCCESS;
	return EXIT_FAILURE;
}

int is_secure_fuse_cst_srk_lock_imx8mp(unsigned char *ocotp_data)
{
	if (secure_status_cst_srk_lock_imx8mp(ocotp_data) == SECURE)
		return 0;
	return 1;
}

enum secure_status secure_status_boot_cfg_lock_imx8mp(unsigned char *ocotp_data)
{
	uint32_t status_mem;
	char *status_str;

	memcpy(&status_mem,
		ocotp_data + IMX8MP_BOOT_CFG_LOCK_OFFSET, 4);

	status_str = boot_cfg_lock_settings_imx8mp(status_mem);

	if (strcmp(status_str, "write protected") == 0)
		return SECURE;

	if (strcmp(status_str, "open") == 0)
		return INSECURE;

	return UNKNOWN;
}

enum secure_status secure_status_cst_srk_lock_imx8mp(unsigned char *ocotp_data)
{
	uint32_t status_mem;
	char *status_str;

	memcpy(&status_mem,
		ocotp_data + IMX8MP_BOOT_CFG_LOCK_OFFSET, 4);

	status_str = cst_srk_lock_settings_imx8mp(status_mem);

	if (strcmp(status_str, "write protected") == 0)
		return SECURE;

	if (strcmp(status_str, "open") == 0)
		return INSECURE;

	return UNKNOWN;
}

int write_fuse_lock_imx8mp(int fd)
{
	uint32_t buf = 0x0;

	buf = IMX8MP_BOOT_CFG_LOCK_OPWP;
	if (pwrite(fd, &buf, sizeof(buf), IMX8MP_BOOT_CFG_LOCK_OFFSET) < 0) {
		fprintf(stderr,
			"Could not write 0x%08X to %s (offset is 0x%02X) : %m\n",
			buf, IMX_OCOTP_NVMEM, IMX8MP_BOOT_CFG_LOCK_OFFSET
		);
		return 1;
	}

	return 0;
}

char *jtag_settings_imx8mp(uint32_t status_mem)
{
	if ((status_mem & IMX8MP_SJC_DISABLE_MASK)
	    == IMX8MP_SJC_DISABLE_TRUE)
		return "disabled";

	if ((status_mem & IMX8MP_JTAG_SMODE_MASK)
	    == IMX8MP_JTAG_SMODE_JTAG_ENABLE)
		return "open";

	switch (status_mem &
		(IMX8MP_JTAG_SMODE_MASK | IMX8MP_KTE_MASK)) {
	case IMX8MP_JTAG_SMODE_SECURE:
		return "secure mode. WARNING: bus tracing is allowed";
	case IMX8MP_JTAG_SMODE_SECURE
	    | IMX8MP_KTE_KILL:
		return "secure mode";
	case IMX8MP_JTAG_SMODE_NO_DEBUG:
		return "bus tracing only allowed";
	case IMX8MP_JTAG_SMODE_NO_DEBUG
	    | IMX8MP_KTE_KILL:
		return "disabled";
	}

	return "unknown";
}

int print_fuse_status_jtag_imx8mp(unsigned char *ocotp_data)
{
	uint32_t status_mem;
	char *status_str;

	memcpy(&status_mem,
		ocotp_data + IMX8MP_JTAG_OFFSET, 4);

	status_str = jtag_settings_imx8mp(status_mem);
	PRINTF_CHECK("jtag %s\n", status_str);

	return 0;
}

enum secure_status secure_status_jtag_imx8mp(unsigned char *ocotp_data)
{
	uint32_t status_mem;
	char *status_str;

	memcpy(&status_mem,
		ocotp_data + IMX8MP_JTAG_OFFSET, 4);

	status_str = jtag_settings_imx8mp(status_mem);

	if ((strcmp(status_str, "secure mode") == 0)
	    || (strcmp(status_str, "bus tracing only allowed") == 0)
	    || (strcmp(status_str, "disabled") == 0))
		return SECURE;

	if (strcmp(status_str, "unknown") == 0)
		return UNKNOWN;

	return INSECURE;
}

int is_secure_fuse_jtag_imx8mp(unsigned char *ocotp_data)
{
	if (secure_status_jtag_imx8mp(ocotp_data) == SECURE)
		return EXIT_SUCCESS;
	return EXIT_FAILURE;
}

int write_fuse_jtag_imx8mp(int fd)
{
	uint32_t buf = 0x0;

	buf = IMX8MP_SJC_DISABLE_TRUE;
	if (pwrite(fd, &buf, sizeof(buf), IMX8MP_JTAG_OFFSET) < 0) {
		fprintf(stderr,
			"Could not write 0x%08X to %s (offset is 0x%02X) : %m\n",
			buf, IMX_OCOTP_NVMEM, IMX8MP_JTAG_OFFSET);
		return 1;
	}

	return 0;
}

char *sd_boot_settings_imx8mp(
	uint32_t bootModeFuses_btFuseSel_mem,
	uint32_t forceBtFromFuse_mem)
{
	if ((bootModeFuses_btFuseSel_mem
	    & IMX8MP_BT_FUSE_SEL_MASK)
	    != IMX8MP_BT_FUSE_SEL_PROGRAMMED) {
		return "allowed";
	}

	if ((forceBtFromFuse_mem
	    & IMX8MP_FORCE_BT_FROM_FUSE_MASK)
	    != IMX8MP_FORCE_BT_FROM_FUSE_ENABLE) {
		return "allowed";
	}

	if ((bootModeFuses_btFuseSel_mem
	    & IMX8MP_BOOT_MODE_FUSES_MASK)
	    != IMX8MP_BOOT_MODE_FUSES_EMMC) {
		return "unknown";
	}

	return "disabled";
}

int print_fuse_status_sd_boot_imx8mp(unsigned char *ocotp_data)
{
	uint32_t bootModeFuses_btFuseSel_mem;
	uint32_t forceBtFromFuse_mem;
	char *status_str;

	memcpy(&bootModeFuses_btFuseSel_mem,
		ocotp_data + IMX8MP_BOOT_MODE_FUSES_OFFSET, 4);

	memcpy(&forceBtFromFuse_mem,
		ocotp_data + IMX8MP_FORCE_BT_FROM_FUSE_OFFSET, 4);

	status_str = sd_boot_settings_imx8mp(
		bootModeFuses_btFuseSel_mem,
		forceBtFromFuse_mem);

	PRINTF_CHECK("sd_boot %s\n", status_str);

	return 0;
}

int write_fuse_sd_boot_imx8mp(int fd)
{
	uint32_t buf = 0x0;

	buf = IMX8MP_BT_FUSE_SEL_PROGRAMMED | IMX8MP_BOOT_MODE_FUSES_EMMC;
	if (pwrite(fd, &buf, sizeof(buf), IMX8MP_BOOT_MODE_FUSES_OFFSET) < 0) {
		fprintf(stderr,
			"Could not write 0x%08X to %s (offset is 0x%02X) : %m\n",
			buf, IMX_OCOTP_NVMEM, IMX8MP_BOOT_MODE_FUSES_OFFSET);
		return 1;
	}

	buf = IMX8MP_FORCE_BT_FROM_FUSE_ENABLE;
	if (pwrite(fd, &buf, sizeof(buf), IMX8MP_FORCE_BT_FROM_FUSE_OFFSET) < 0) {
		fprintf(stderr,
			"Could not write 0x%08X to %s (offset is 0x%02X) : %m\n",
			buf, IMX_OCOTP_NVMEM, IMX8MP_FORCE_BT_FROM_FUSE_OFFSET);
		return 1;
	}

	return 0;
}

enum secure_status secure_status_sd_boot_imx8mp(unsigned char *ocotp_data)
{
	uint32_t bootModeFuses_btFuseSel_mem;
	uint32_t forceBtFromFuse_mem;
	char *status_str;

	memcpy(&bootModeFuses_btFuseSel_mem,
		ocotp_data + IMX8MP_BOOT_MODE_FUSES_OFFSET, 4);

	memcpy(&forceBtFromFuse_mem,
		ocotp_data + IMX8MP_FORCE_BT_FROM_FUSE_OFFSET, 4);

	status_str = sd_boot_settings_imx8mp(
		bootModeFuses_btFuseSel_mem,
		forceBtFromFuse_mem);

	if (strcmp(status_str, "disabled") == 0)
		return SECURE;

	if (strcmp(status_str, "unknown") == 0)
		return UNKNOWN;

	return INSECURE;
}

int is_secure_fuse_sd_boot_imx8mp(unsigned char *ocotp_data)
{
	if (secure_status_sd_boot_imx8mp(ocotp_data) == SECURE)
		return EXIT_SUCCESS;
	return EXIT_FAILURE;
}

char *secureboot_settings_imx8mp(
	uint32_t *hash_mem,
	uint32_t sec_config_mem)
{
	int i;

	for (i = 0; i < 8; i++) {
		if (hash_mem[i] != 0)
			break;
		if (i == 7)
			return "disabled";
	}

	if ((sec_config_mem
	    & IMX8MP_SEC_CONFIG_MASK)
	    != IMX8MP_SEC_CONFIG_ENABLE) {
		return "configured (not enforced)";
	}

	return "configured";
}

int print_fuse_status_secureboot_imx8mp(unsigned char *ocotp_data)
{
	int i;
	uint32_t hash_mem[8];
	uint32_t sec_config_mem;
	char *status_str;

	// value of SRK region.
	for (i = 0; i < 8; i++)
		memcpy(&hash_mem[i],
			ocotp_data + IMX8MP_SRK_HASH_OFFSET + 4*i, 4);

	// close process. If the signature confirmation fails, the startup will fail.
	memcpy(&sec_config_mem,
			ocotp_data + IMX8MP_SEC_CONFIG_OFFSET, 4);

	status_str = secureboot_settings_imx8mp(
		hash_mem,
		sec_config_mem);

	PRINTF_CHECK("secureboot %s\n", status_str);

	return 0;
}

enum secure_status secure_status_secureboot_imx8mp(unsigned char *ocotp_data)
{
	int i;
	uint32_t hash_mem[8];
	uint32_t sec_config_mem;
	char *status_str;

	// value of SRK region.
	for (i = 0; i < 8; i++)
		memcpy(&hash_mem[i],
			ocotp_data + IMX8MP_SRK_HASH_OFFSET + 4*i, 4);

	// close process. If the signature confirmation fails, the startup will fail.
	memcpy(&sec_config_mem,
			ocotp_data + IMX8MP_SEC_CONFIG_OFFSET, 4);

	status_str = secureboot_settings_imx8mp(
		hash_mem,
		sec_config_mem);

	if (strcmp(status_str, "configured") == 0)
		return SECURE;

	if (strcmp(status_str, "disabled") == 0)
		return INSECURE;

	return UNKNOWN;
}

int write_fuse_secureboot_srk_imx8mp(int fd)
{
	int i;
	uint32_t buf = 0x0;

	// value of SRK region.
	for (i = 0; i < 8; i++) {
		if (pwrite(fd, &srk_hashes[i], sizeof(srk_hashes[i]),
		    IMX8MP_SRK_HASH_OFFSET + 4*i) < 0) {
			fprintf(stderr,
				"Could not write 0x%08X to %s (offset is 0x%02X, SRK%d) : %m\n",
				srk_hashes[i], IMX_OCOTP_NVMEM, IMX8MP_SRK_HASH_OFFSET + 4*i, i
			);
			return 1;
		}
	}

	// cst_srk_lock.
	buf = IMX8MP_CST_SRK_LOCK_ENABLE;
	if (pwrite(fd, &buf, 4, IMX8MP_BOOT_CFG_LOCK_OFFSET) < 0) {
		fprintf(stderr,
			"Could not write 0x%08X to %s (offset is 0x%02X, SEC_CONFIG) : %m\n",
			buf, IMX_OCOTP_NVMEM, IMX8MP_BOOT_CFG_LOCK_OFFSET
		);
		return 1;
	}

	return 0;
}

int write_fuse_secureboot_close_imx8mp(int fd)
{
	uint32_t buf = 0x0;

	// close.
	buf = IMX8MP_SEC_CONFIG_ENABLE;
	if (pwrite(fd, &buf, 4, IMX8MP_SEC_CONFIG_OFFSET) < 0) {
		fprintf(stderr,
			"Could not write 0x%08X to %s (offset is 0x%02X, SEC_CONFIG) : %m\n",
			buf, IMX_OCOTP_NVMEM, IMX8MP_SEC_CONFIG_OFFSET
		);
		return 1;
	}

	return 0;
}

int cmp_fuse_secureboot_srk_imx8mp(int fd)
{
	int i;
	uint32_t buf = 0x0;

	for (i = 0; i < 8; i++) {
		if (pread(fd, &buf, sizeof(buf), IMX8MP_SRK_HASH_OFFSET + 4*i) < 0) {
			fprintf(stderr,
				"Could not read %s (offset is 0x%02X, SRK%d) : %m\n",
				IMX_OCOTP_NVMEM, IMX8MP_SRK_HASH_OFFSET + 4*i, i
			);
			return 1;
		}
		if (srk_hashes[i] != buf)
			return 1;
	}

	return 0;
}

int show_srk_imx8mp(int fd)
{
	int i;
	uint32_t buf = 0x0;

	for (i = 0; i < 8; i++) {
		if (pread(fd, &buf, sizeof(buf), IMX8MP_SRK_HASH_OFFSET + 4*i) < 0) {
			fprintf(stderr,
				"Could not read %s (offset is 0x%02X, SRK%d) : %m\n",
				IMX_OCOTP_NVMEM, IMX8MP_SRK_HASH_OFFSET + 4*i, i
			);
			return 1;
		}

		PRINTF_CHECK("0x%08X", buf);

		if (i < 7)
			PRINTF_CHECK(",");
	}

	PRINTF_CHECK("\n");

	return 0;
}

int is_secure_fuse_secureboot_imx8mp(unsigned char *ocotp_data)
{
	if (secure_status_secureboot_imx8mp(ocotp_data) == SECURE)
		return EXIT_SUCCESS;
	return EXIT_FAILURE;
}
