// SPDX-License-Identifier: MIT
/*
 * Copyright (c) 2023 Atmark Techno,Inc.
 */
#include <assert.h>
#include "../security_imx8mp.h"

int test_boot_cfg_lock_settings_imx8mp(void)
{
	char *status_char;
	uint32_t status_mem;

	status_mem = IMX8MP_BOOT_CFG_LOCK_OPWP;
	status_char = boot_cfg_lock_settings_imx8mp(status_mem);
	assert((strcmp(status_char, "write protected") == 0));

	status_mem = 0x8;
	status_char = boot_cfg_lock_settings_imx8mp(status_mem);
	assert((strcmp(status_char, "open") == 0));

	return 0;
}

int test_cst_srk_lock_settings_imx8mp(void)
{
	char *status_char;
	uint32_t status_mem;

	status_mem = IMX8MP_CST_SRK_LOCK_ENABLE;
	status_char = cst_srk_lock_settings_imx8mp(status_mem);
	assert((strcmp(status_char, "write protected") == 0));

	status_mem = 0x0;
	status_char = cst_srk_lock_settings_imx8mp(status_mem);
	assert((strcmp(status_char, "open") == 0));

	return 0;
}

int test_jtag_settings_imx8mp(void)
{
	char *status_char;
	uint32_t status_mem;

	status_mem = IMX8MP_SJC_DISABLE_TRUE
	    | IMX8MP_JTAG_SMODE_SECURE
	    | IMX8MP_KTE_KILL;
	status_char = jtag_settings_imx8mp(status_mem);
	assert((strcmp(status_char, "disabled") == 0));

	status_mem = IMX8MP_JTAG_SMODE_JTAG_ENABLE | IMX8MP_KTE_KILL;
	status_char = jtag_settings_imx8mp(status_mem);
	assert((strcmp(status_char, "open") == 0));

	// JTAG_SMODE is "reserved" mode.
	status_mem = 0x800000 | IMX8MP_KTE_KILL;
	status_char = jtag_settings_imx8mp(status_mem);
	assert((strcmp(status_char, "unknown") == 0));

	status_mem = IMX8MP_JTAG_SMODE_SECURE;
	status_char = jtag_settings_imx8mp(status_mem);
	assert((strcmp(status_char, "secure mode. WARNING: bus tracing is allowed") == 0));

	status_mem = IMX8MP_JTAG_SMODE_SECURE | IMX8MP_KTE_KILL;
	status_char = jtag_settings_imx8mp(status_mem);
	assert((strcmp(status_char, "secure mode") == 0));

	status_mem = IMX8MP_JTAG_SMODE_NO_DEBUG;
	status_char = jtag_settings_imx8mp(status_mem);
	assert((strcmp(status_char, "bus tracing only allowed") == 0));

	status_mem = IMX8MP_JTAG_SMODE_NO_DEBUG | IMX8MP_KTE_KILL;
	status_char = jtag_settings_imx8mp(status_mem);
	assert((strcmp(status_char, "disabled") == 0));

	return 0;
}

int test_sd_boot_settings_imx8mp(void)
{
	char *status_char;
	uint32_t bootModeFuses_btFuseSel_mem;
	uint32_t forceBtFromFuse_mem;

	bootModeFuses_btFuseSel_mem = IMX8MP_BOOT_MODE_FUSES_EMMC;
	forceBtFromFuse_mem = IMX8MP_FORCE_BT_FROM_FUSE_ENABLE;

	status_char = sd_boot_settings_imx8mp(
		bootModeFuses_btFuseSel_mem,
		forceBtFromFuse_mem);

	assert((strcmp(status_char, "allowed") == 0));

	bootModeFuses_btFuseSel_mem = IMX8MP_BT_FUSE_SEL_PROGRAMMED
		| IMX8MP_BOOT_MODE_FUSES_EMMC;
	forceBtFromFuse_mem = 0x0;

	status_char = sd_boot_settings_imx8mp(
		bootModeFuses_btFuseSel_mem,
		forceBtFromFuse_mem);

	assert((strcmp(status_char, "allowed") == 0));

	bootModeFuses_btFuseSel_mem = IMX8MP_BT_FUSE_SEL_PROGRAMMED;
	forceBtFromFuse_mem = IMX8MP_FORCE_BT_FROM_FUSE_ENABLE;

	status_char = sd_boot_settings_imx8mp(
		bootModeFuses_btFuseSel_mem,
		forceBtFromFuse_mem);

	assert((strcmp(status_char, "unknown") == 0));

	bootModeFuses_btFuseSel_mem = IMX8MP_BT_FUSE_SEL_PROGRAMMED
		| IMX8MP_BOOT_MODE_FUSES_EMMC;
	forceBtFromFuse_mem = IMX8MP_FORCE_BT_FROM_FUSE_ENABLE;

	status_char = sd_boot_settings_imx8mp(
		bootModeFuses_btFuseSel_mem,
		forceBtFromFuse_mem);

	assert((strcmp(status_char, "disabled") == 0));


	return 0;
}

int test_secureboot_settings_imx8mp(void)
{
	char *status_char;
	uint32_t hash_mem[8];
	uint32_t sec_config_mem;
	int i;

	for (i = 0; i < 8; i++)
		hash_mem[i] = 0x0;
	sec_config_mem = IMX8MP_SEC_CONFIG_ENABLE;

	status_char = secureboot_settings_imx8mp(
		hash_mem,
		sec_config_mem);

	assert((strcmp(status_char, "disabled") == 0));

	for (i = 0; i < 8; i++)
		hash_mem[i] = 0x1;
	sec_config_mem = 0x0;

	status_char = secureboot_settings_imx8mp(
		hash_mem,
		sec_config_mem);

	assert((strcmp(status_char, "configured (not enforced)") == 0));

	for (i = 0; i < 8; i++)
		hash_mem[i] = 0x1;
	sec_config_mem = IMX8MP_SEC_CONFIG_ENABLE;

	status_char = secureboot_settings_imx8mp(
		hash_mem,
		sec_config_mem);

	assert((strcmp(status_char, "configured") == 0));

	return 0;
}

int main(void)
{
	printf("-------- test security_imx8mp ---------\n");
	test_boot_cfg_lock_settings_imx8mp();
	test_cst_srk_lock_settings_imx8mp();
	printf("lock: ok\n");
	test_jtag_settings_imx8mp();
	printf("jtag: ok\n");
	test_sd_boot_settings_imx8mp();
	printf("sd_boot: ok\n");
	test_secureboot_settings_imx8mp();
	printf("secureboot: ok\n");

	return 0;
}
